import json


__all__ = ('Error', 'RequestError')


class Error(Exception):

    """
    Raised when something flops.
    """

    __slots__ = ()


class RequestError(Error):

    """
    Raised when a request fails.
    """

    __slots__ = ('response', 'data')

    def __init__(self, response, data):

        message = json.dumps(data, indent = 4)

        super().__init__(message)

        self.response = response

        self.data = data



import functools

from . import helpers


__all__ = ('missing', 'Structure')


missing = type('missing', (), {'__slots__': (), '__bool__': lambda s: False})()


class Structure:

    __slots__ = ('__dict__',)

    def __init__(self, *args, **extra):

        try:

            (data,) = args

        except ValueError:

            pass

        else:

            extra.update(data)

        update_structure(self, extra)

    def __getattr__(self, name, forbidden = '__'):

        if name.startswith(forbidden):

            raise AttributeError(name)

        blueprint = _blueprint(self.__class__)

        available = blueprint.keys()

        if name in available:

            return missing

        raise AttributeError(name)

    def __eq__(self, other):

        family = isinstance(other, self.__class__)

        return family and self.__dict__ == other.__dict__

    def __ne__(self, other):

        return not self.__eq__(other)

    def __hash__(self):

        return super().__hash__()

    def __repr__(self):

        return f'<{self.__class__.__name__}>'


@functools.lru_cache(maxsize = None)
def _structure(name):

    bases = (Structure,)

    namespace = {}

    value = type(name, bases, namespace)

    return value


@functools.lru_cache(maxsize = None)
def _blueprint(structure):

    return _blueprints[structure.__name__]


@functools.lru_cache(maxsize = None)
def _structure_update():

    skip = (missing,)

    def wrapper(value, data):

        blueprint = _blueprint(value.__class__)

        return helpers.update_generic(blueprint, value, data, skip = skip)

    return wrapper


def update_structure(value, data):

    return _structure_update()(value, data)


def _noop_update(value, data):

    return


def _pick_update(value):

    if isinstance(value, type) and issubclass(value, Structure):

        return _structure_update()

    return _noop_update


@functools.lru_cache(maxsize = None)
def _list_update(build):

    update = _pick_update(build)

    return helpers.cached_partial(helpers.update_list, update, build)


def update_list(build, *args, **kwargs):

    return _list_update(build)(*args, **kwargs)


@functools.lru_cache(maxsize = None)
def _dict_update(build, keys):

    update = _pick_update(build)

    identify = helpers.cached_partial(helpers.crawl, keys)

    return helpers.cached_partial(helpers.update_dict, update, build, identify)


def update_dict(build, keys, *args, **kwargs):

    return _dict_update(build, keys)(*args, **kwargs)


@functools.lru_cache(maxsize = None)
def _field(create, update):

    return (create, update)


_string = str


_integer = int


_boolean = bool


_list = list


_dict = dict



_blueprints = {
    'Track': {
        'id': _field(
            _integer,
            None
        ),
        'created_at': _field(
            _string,
            None
        ),
        'user_id': _field(
            _integer,
            None
        ),
        'duration': _field(
            _integer,
            None
        ),
        'commentable': _field(
            _boolean,
            None
        ),
        'state': _field(
            _string,
            None
        ),
        'sharing': _field(
            _string,
            None
        ),
        'tag_list': _field(
            _string,
            None
        ),
        'permalink': _field(
            _string,
            None
        ),
        'description': _field(
            _string,
            None
        ),
        'downloadable': _field(
            _boolean,
            None
        ),
        'genre': _field(
            _string,
            None
        ),
        'release': _field(
            _string,
            None
        ),
        'purchase_url': _field(
            _string,
            None
        ),
        'label_id': _field(
            _integer,
            None
        ),
        'label_name': _field(
            _string,
            None
        ),
        'isrc': _field(
            _string,
            None
        ),
        'video_url': _field(
            _string,
            None
        ),
        'track_type': _field(
            _string,
            None
        ),
        'key_signature': _field(
            _string,
            None
        ),
        'bpm': _field(
            _integer,
            None
        ),
        'title': _field(
            _string,
            None
        ),
        'release_year': _field(
            _integer,
            None
        ),
        'release_day': _field(
            _integer,
            None
        ),
        'release_month': _field(
            _string,
            None
        ),
        'original_format': _field(
            _string,
            None
        ),
        'original_content_size': _field(
            _integer,
            None
        ),
        'licence': _field(
            _string,
            None
        ),
        'uri': _field(
            _string,
            None
        ),
        'permalink_url': _field(
            _string,
            None
        ),
        'artwork_url': _field(
            _string,
            None
        ),
        'waveform_url': _field(
            _string,
            None
        ),
        'user': _field(
            _structure('User'),
            _structure_update()
        ),
        'stream_url': _field(
            _string,
            None
        ),
        'download_url': _field(
            _string,
            None
        ),
        'playback_count': _field(
            _integer,
            None
        ),
        'download_count': _field(
            _integer,
            None
        ),
        'favoritings_count': _field(
            _integer,
            None
        ),
        'comment_count': _field(
            _integer,
            None
        ),
        'attachments_uri': _field(
            _string,
            None
        ),
    },
    'User': {
        'id': _field(
            _integer,
            None
        ),
        'permalink': _field(
            _string,
            None
        ),
        'username': _field(
            _string,
            None
        ),
        'uri': _field(
            _string,
            None
        ),
        'permalink_url': _field(
            _string,
            None
        ),
        'avatar_url': _field(
            _string,
            None
        )
    },
    'Playlist': {
        'id': _field(
            _integer,
            None
        ),
        'created_at': _field(
            _string,
            None
        ),
        'user_id': _field(
            _integer,
            None
        ),
        'user': _field(
            _structure('User'),
            _structure_update()
        ),
        'tracks': _field(
            _list,
            _list_update(
                _structure('Track')
            )
        ),
        'title': _field(
            _string,
            None
        ),
        'permalink': _field(
            _string,
            None
        ),
        'permalink_url': _field(
            _string,
            None
        ),
        'uri': _field(
            _string,
            None
        ),
        'sharing': _field(
            _string,
            None
        ),
        'embeddable_by': _field(
            _string,
            None
        ),
        'purchase_url': _field(
            _string,
            None
        ),
        'artwork_url': _field(
            _string,
            None
        ),
        'description': _field(
            _string,
            None
        ),
        'label': _field(
            _structure('User'),
            _structure_update()
        ),
        'duration': _field(
            _integer,
            None
        ),
        'genre': _field(
            _string,
            None
        ),
        'tag_list': _field(
            _string,
            None
        ),
        'label_id': _field(
            _integer,
            None
        ),
        'label_name': _field(
            _string,
            None
        ),
        'release': _field(
            _string,
            None
        ),
        'release_day': _field(
            _integer,
            None
        ),
        'release_month': _field(
            _integer,
            None
        ),
        'release_year': _field(
            _integer,
            None
        ),
        'streamable': _field(
            _boolean,
            None
        ),
        'downloable': _field(
            _boolean,
            None
        ),
        'ean': _field(
            _string,
            None
        ),
        'playlist_type': _field(
            _string,
            None
        )
    }
}


@functools.lru_cache(maxsize = None)
def __getattr__(name):

    available = _blueprints.keys()

    if not name in available:

        raise AttributeError(name)

    return _structure(name)


__all__ = (*__all__, *_blueprints.keys())



import asyncio
import aiohttp
import yarl
import json

from . import errors
from . import storage
from . import helpers


__all__ = ('Client',)


class Client:

    _base = yarl.URL('https://api.soundcloud.com')

    def __init__(self, session, client_id):

        self._client_id = client_id

        self._session = session

    async def request(self, method, url, **kwargs):

        params = helpers.get_soft(kwargs, 'params', dict)

        helpers.fix_params(params)

        params['client_id'] = self._client_id

        response = await self._session.request(method, url, **kwargs)

        return response

    async def touch(self, method, path, **kwargs):

        url = self._base.with_path(self._base.path + path)

        response = await self.request(method, url, **kwargs)

        data = await response.text()

        if not response.status == 204:

            data = json.loads(data)

        if response.status < 400:

            return data

        raise errors.RequestError(response, data)

    _resolves = {
        'user': storage.User,
        'track': storage.Track,
        'playlist': storage.Playlist
    }

    async def resolve(self, value):

        params = {'url': value}

        data = await self.touch('GET', 'resolve', params = params)

        kind = data.pop('kind')

        structure = self._resolves[kind]

        value = structure(data)

        return value

    async def stream(self, value):

        response = await self.request('GET', value)

        length = int(response.headers['Content-Length'])

        return (response.content, length)



import functools


__all__ = ()


def get_soft(root, key, factory):

    try:

        value = root[key]

    except KeyError:

        value = root[key] = factory()

    return value


def fix_params(root, booleans = ('false', 'true')):

    for (key, value) in tuple(root.items()):

        if value is None:

            del root[key]

            continue

        if isinstance(value, bool):

            value = _booleans[value]

        elif isinstance(value, (float, int)):

            value = str(value)

        else:

            continue

        root[key] = value


def update_generic(blueprint, value, datas, skip = ()):

    for (key, data) in datas.items():

        try:

            (create, update) = blueprint[key]

        except KeyError:

            handle = False

        else:

            handle = not data is None

        if handle:

            if update is None:

                new = create(data)

            else:

                try:

                    new = getattr(value, key)

                except AttributeError:

                    missing = True

                else:

                    missing = new in skip

                if missing or new is None:

                    new = create()

                update(new, data)

                if not missing:

                    continue
        else:

            new = data

        setattr(value, key, new)


def update_list(update, create, value, datas):

    keep = []

    for data in datas:

        contest = create(data)

        for inner in value:

            if not inner == contest:

                continue

            if not update is None:

                update(inner, data)

            break

        else:

            inner = contest

            value.append(inner)

        keep.append(inner)

    index = 0

    while index < len(value):

        inner = value[index]

        if inner in keep:

            index += 1

        else:

            value.remove(inner)


def update_dict(update, create, identify, value, datas, clean = True):

    fresh = []

    for data in datas:

        identity = identify(data)

        try:

            inner = value[identity]

        except KeyError:

            inner = value[identity] = create()

        update(inner, data)

        fresh.append(inner)

    if not clean:

        return

    for key, inner in tuple(value.items()):

        if inner in fresh:

            continue

        del value[key]


@functools.lru_cache(maxsize = None)
def cached_partial(*args, **kwargs):

    return functools.partial(*args, **kwargs)


def crawl(keys, data):

    for key in keys:

        data = data[key]

    return data
